library crm.greentea.fe3.widgets.charts.src.chart_annotation;

import 'dart:html' show Element;
import 'dart:svg' show LineElement;

import 'package:charted/charted.dart';
import 'package:meta/meta.dart';

typedef void OnAnnotationDismissClicked(ChartEvent e);

/// Creates annotation for a line chart.  The annotation will be displayed above
/// the chart area using absolutely positioned divs much like a hovercard.  The
/// annotation has a fixed width but can grow as tall as it needs to be (no
/// max-height).  The annotation should generally be only two to three lines of
/// text.
class ChartAnnotation implements ChartBehavior {
  // Offset to the top of the chart.
  static const _ANNOTATION_PADDING = 8;
  static const _ANNOTATION_EDGE_LENGTH = 10;
  static const _BASE_ZINDEX = 900;
  static const _HOVER_ZINDEX = 1000;

  final int _annotationColumnIndex;
  final int _colorColumnIndex;
  final OnAnnotationDismissClicked _dismissCallBack;

  // The list of all annotations, shown or hidden.
  List<Element> _annotations = [];

  // The list of hidden annotations.
  List<Element> _hiddenAnnotations = [];

  // The annotation and the grid line are two separate element but needs to be
  // added and removed together.  Keep track of them by this map.
  Map<Element, Element> _annotationGridMap = {};
  CartesianArea _area;
  SelectionScope _scope;
  Element _upperRenderPane;

  // The area above the ChartArea to allow absolute positioning annotations
  // with "bottom" style to allow growing the div upwards.  All annotations are
  // added into this area.
  Element _annotationArea;
  SubscriptionsDisposer _disposer = new SubscriptionsDisposer();

  /// The [annotationColumnIndex] is the index of the column in the ChartData
  /// for the ChartAnnotation to read the available annotations.
  /// If [colorColumnIndex] is set, it will be the index of the column in the
  /// ChartData for the ChartAnnotation to read the color for the annotation.
  /// [dismissCallBack] is called when the 'X' button is clicked on the
  /// annotation
  ChartAnnotation(
      {@required annotationColumnIndex,
      int colorColumnIndex,
      OnAnnotationDismissClicked dismissCallBack})
      : this._annotationColumnIndex = annotationColumnIndex,
        this._colorColumnIndex = colorColumnIndex,
        this._dismissCallBack = dismissCallBack;

  void init(ChartArea area, Selection upperRenderPane, Selection __) {
    _area = area;
    _scope = new SelectionScope.element(_area.host);
    _upperRenderPane = upperRenderPane.first;

    // Draw the annotation once axes are initialized, because we rely on the
    // scale to compute position.
    _area.onChartAxesUpdated.listen((_) => _drawAnnotations());
  }

  void _drawAnnotations() {
    _clearAll();
    var rect = _area.layout.renderArea;
    var dimensionScale = _area.dimensionScales.first;
    var dimIndex = _area.config.dimensions.first;

    _annotationArea = new Element.div();
    _annotationArea.style
      ..height = '1px'
      ..position = 'absolute'
      ..top = (rect.y + _ANNOTATION_EDGE_LENGTH / 2).toString() + 'px';
    _area.host.style.position = 'relative';
    _area.host.append(_annotationArea);

    var rows = _area.data.rows;
    var annotationCount = 0;
    for (var i = 0; i < rows.length; i++) {
      var row = rows.elementAt(i);
      String text = row[_annotationColumnIndex];
      if (text == null || text.isEmpty) continue;
      annotationCount++;

      // Gets the xposition of the row with annotation and draw a vertical line
      var xPos = dimensionScale.scale(row[dimIndex]);
      var annotationColor =
          _colorColumnIndex == null ? 'grey' : row[_colorColumnIndex];

      // Create the annotation box.
      var annotation = new Element.div();
      annotation.text = annotationCount.toString();
      annotation.classes = ['annotation'];
      annotation.style
        ..bottom = (rect.y - _ANNOTATION_EDGE_LENGTH).toString() + 'px'
        ..left =
            (rect.x + xPos - _ANNOTATION_EDGE_LENGTH / 2 - _ANNOTATION_PADDING)
                    .toString() +
                'px'
        ..padding = '${_ANNOTATION_PADDING}px ${_ANNOTATION_PADDING}px'
        ..width = '${_ANNOTATION_EDGE_LENGTH}px'
        ..border = '1px solid $annotationColor'
        ..color = annotationColor
        ..zIndex = (_BASE_ZINDEX + i).toString();

      // Keeping the z index is important because if user hides then restore the
      // annotation, they will stack up in the correct order.
      annotation.dataset = {'zIndex': (_BASE_ZINDEX + i).toString()};

      var dismissButton = new Element.div();
      dismissButton.classes = ['dismiss-button'];
      dismissButton.text = 'X';
      dismissButton.style.opacity = '0';
      _disposer.add(dismissButton.onClick.listen((mouseEvent) {
        // Hide annotation, restore z index.
        annotation.style..zIndex = annotation.dataset['zIndex'];
        dismissButton.style.opacity = '0';
        annotation.remove();
        annotation.classes.remove('annotation-hover');
        _hiddenAnnotations.add(annotation);

        // Call the dismiss callback if it's registered.
        if (_dismissCallBack != null) {
          var chartEvent = new ChartEvent(
              mouseEvent, _area, null, i, _annotationColumnIndex);
          _dismissCallBack(chartEvent);
        }
      }));

      // Apply hover style and hovered z index.
      _disposer.add(annotation.onMouseOver.listen((mouseEvent) {
        dismissButton.style.opacity = '1';
        annotation.style..zIndex = _HOVER_ZINDEX.toString();
        annotation.classes.add('annotation-hover');
      }));

      // Restore z index, remove hover style
      _disposer.add(annotation.onMouseOut.listen((mouseEvent) {
        dismissButton.style.opacity = '0';
        annotation.style..zIndex = annotation.dataset['zIndex'];
        annotation.classes.remove('annotation-hover');
      }));

      annotation.append(dismissButton);
      _annotations.add(annotation);
      _annotationArea.append(annotation);
    }
  }

  /// Shows all annotations.
  void showAll() {
    for (var annotation in _hiddenAnnotations) {
      _annotationArea.append(annotation);
      _upperRenderPane.append(_annotationGridMap[annotation]);
    }
    _hiddenAnnotations.clear();
  }

  /// Hides all annotations.
  void hideAll() {
    for (var annotation in _annotations) {
      _hiddenAnnotations.add(annotation);

      annotation.remove();
      _annotationGridMap[annotation].remove();
    }
  }

  /// Clear all annotations.
  void _clearAll() {
    for (var annotation in _annotations) {
      annotation.remove();
      _annotationGridMap[annotation].remove();
    }
    _annotations.clear();
    _annotationGridMap.clear();
    _hiddenAnnotations.clear();
  }

  /// Show all hidden annotation of a given color.
  void showAllByColor(String color) {
    var shownAnnotations = [];
    for (var annotation in _hiddenAnnotations) {
      if (annotation.style.color == color) {
        _annotationArea.append(annotation);
        _upperRenderPane.append(_annotationGridMap[annotation]);
        shownAnnotations.add(annotation);
      }
    }
    shownAnnotations.forEach((e) => _hiddenAnnotations.remove(e));
  }

  void dispose() {
    _disposer.dispose();
    _annotations.forEach((e) => e.remove());
  }
}
